﻿import gym
import numpy as np
import chainer
import chainer.functions as F
import chainer.links as L
import chainerrl

class QFunction(chainer.Chain):
    def __init__(self, obs_size, hidden, n_actions):
        super().__init__()
        with self.init_scope():
            self.l0 = L.Linear(obs_size, hidden)                                   # 入力数,中間層のノード
            self.l1 = L.Linear(hidden, hidden)                                     # 中間層,中間層のノード        
            self.l2 = L.Linear(hidden, n_actions)                                  # 中間層,出力層のノード
    def __call__(self, x, test=False):
        h0 = F.tanh(self.l0(x))                                                    # 中間層の活性化関数
        h1 = F.tanh(self.l1(h0))                                                   # 中間層の活性化関数
        h2 = self.l2(h1)                                                           # 出力層の活性化関数(恒等関数)
        return chainerrl.action_value.DiscreteActionValue(h2)                      # 深層強化学習関数

env        = gym.make('CartPole-v0')                                               # 倒立振子
steps      = 200                                                                   # 1試行のstep数
n_episodes = 400                                                                   # 総試行回数
gamma      = 0.99

# DQN設定
q_func     = QFunction(env.observation_space.shape[0], 50, env.action_space.n)     # 入力数4, 中間層50, 出力層2
opt        = chainer.optimizers.Adam(eps=1e-2)                                     # 最適化関数
opt.setup(q_func)
explorer   = chainerrl.explorers.LinearDecayEpsilonGreedy \
             (start_epsilon=1, end_epsilon=0.1, decay_steps=n_episodes, random_action_func=env.action_space.sample)                # ε-greedy法
ex_rep     = chainerrl.replay_buffer.ReplayBuffer(capacity=10 ** 6)                # 経験再生(experience replay)
phi        = lambda x: x.astype(np.float32, copy=False)
agent      = chainerrl.agents.DQN \
             (q_func, opt, ex_rep, gamma, explorer, replay_start_size=500, update_interval=1, target_update_interval=100, phi=phi) # 深層強化学習

for episode in range(n_episodes):                                                  # エピソード数のループ
    observ      = env.reset()
    done        = False
    reward      = 0
    reward_sum  = 0
    for t in range(steps):                                                         # 1試行のループ
        action                   = agent.act_and_train(observ, reward)             # アクション決定
        observ, reward, done, _  = env.step(action)                                # アクション後の観測値
        reward_sum              += reward                                          # 報酬追加
        if done: break
    agent.stop_episode_and_train(observ, reward, done)                             # DQNの重み更新
    if episode % 10 == 0: print('episode:', episode, 'reward_sum:', reward_sum)
agent.save('agent')
